# This code has been developed under THALIS: MATENVMED project for Action 4.3 and co-financed by Greece and EU - NSRF 2007-2013. 
# FPDE python module solves the flow potential problem for Kalymnos aquifer problem
from dolfin import *
import numpy as np
import scipy as sc
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
  
def FPDE(Q_k,k_values,flagplot,fplot,KK,G1x,G1y,G2x,G2y,G3x,G3y,G4x,G4y,G5x,G5y,G1,G2,G3,G4,G5,iflag,ierror):
    # Mesh definition
    mesh = RectangleMesh(Point(0.0, 0.0),Point(7000.0, 3000.0), 700, 300, "right")
    parameters["reorder_dofs_serial"] = False

    # Delta function for each well
    class Delta(Expression):
        def __init__(self, eps):
            self.eps = eps
        def eval(self, values, x):
            eps = self.eps
            area=100.0
            if x[0]==2640.0 and x[1]==1560.0 :
                values[0] = eps-Q_k[0]/area
            elif x[0]==3340.0 and x[1]==2200.0 :
                values[0] = eps-Q_k[1]/area
            elif x[0]==3920.0 and x[1]==960.0 :
                values[0] = eps-Q_k[2]/area
            elif x[0]==4620.0 and x[1]==2460.0 :
                values[0] = eps-Q_k[3]/area
            elif x[0]==4860.0 and x[1]==1580.0 :
                values[0] = eps-Q_k[4]/area
            else:
                values[0] = eps

    class Left(SubDomain):
        def inside(self, x, on_boundary):
            return near(x[0], 0.0)
    
    class Right(SubDomain):
        def inside(self, x, on_boundary):
            return near(x[0], 7000.0)
    
    class Bottom(SubDomain):
        def inside(self, x, on_boundary):
            return near(x[1], 0.0)
    
    class Top(SubDomain):
        def inside(self, x, on_boundary):
            return near(x[1], 3000.0)
    # Initialize sub-domain instances
    left = Left()
    top = Top()
    right = Right()
    bottom = Bottom()
    
    # Initialize mesh function for interior domains
    domains = CellFunction("size_t", mesh)
    domains.set_all(0)

    class RightUp(SubDomain):
        def inside(self, x, on_boundary):
            tol = DOLFIN_EPS
            return on_boundary and  x[1] >= 1950 and abs(x[0]-7000)<tol
    class RightDown(SubDomain):
        def inside(self, x, on_boundary):
            tol = DOLFIN_EPS
            return on_boundary and  x[1] <= 1950 and abs(x[0]-7000)<tol
    
    # Initialize mesh function for boundary domains
    boundaries = FacetFunction("size_t", mesh)
    boundaries.set_all(0)
    left.mark(boundaries, 1)
    top.mark(boundaries, 2)
    right.mark(boundaries, 3)
    bottom.mark(boundaries, 4)
    rtb=RightUp()
    rdb=RightDown()
    rtb.mark(boundaries, 5)
    rdb.mark(boundaries, 6)

    V0 = FunctionSpace(mesh, 'DG', 0)
    
    kapa = Function(V0)
    kapa.vector()[:]=KK[:]
    
    V = FunctionSpace(mesh, "CG", 1)
    u = TrialFunction(V)
    v = TestFunction(V)
    
    # Define Dirichlet boundary conditions at left and right boundaries
    bcs = [DirichletBC(V, 0.0, boundaries, 1)]
    
    # Define new measures associated with the interior domains and
    # exterior boundaries
    dx = Measure("dx")[domains]
    ds = Measure("ds")[boundaries]

    # The Neumann condition values
    nt=1.23
    nd=1.23
    g_rt = Constant(nt)
    g_rd = Constant(nd)

    g_d = Constant("0.0")
    
    delta= Delta(0.03/365.0)
    # Variational formulation
    a = inner(kapa*grad(u), grad(v))*dx(0)
    L = inner(delta, v)*dx(0) + g_rt*v*ds(5)+g_rd*v*ds(6)
    u = Function(V)
    solve(a == L, u, bcs)

#    plot(u)
#    interactive()

  
    u_nodal_values = u.vector()
    u_array = u_nodal_values.array()
#------------------------------------------------------------------

    phi = 8.0078125
    
    DW=np.zeros((5))
    e1=np.zeros(11, dtype=np.float64)
    e2=np.zeros(11, dtype=np.float64)
    e3=np.zeros(11, dtype=np.float64)
    e4=np.zeros(11, dtype=np.float64)
    e5=np.zeros(11, dtype=np.float64)
    e1=u_array[G1]
    e2=u_array[G2]
    e3=u_array[G3]
    e4=u_array[G4]
    e5=u_array[G5]
    # Evaluation at Guard points
    for i in range(11):
        e1[i]=e1[i]-phi
        e2[i]=e2[i]-phi
        e3[i]=e3[i]-phi
        e4[i]=e4[i]-phi
        e5[i]=e5[i]-phi
    # Activation at Guard points
    if np.min(e1)<DOLFIN_EPS:
        DW[0]=1
    if np.min(e2)<DOLFIN_EPS:
        DW[1]=1
    if np.min(e3)<DOLFIN_EPS:
        DW[2]=1
    if np.min(e4)<DOLFIN_EPS:
        DW[3]=1
    if np.min(e5)<DOLFIN_EPS:
        DW[4]=1
    coor = mesh.coordinates()

    k1=0
    # Define salt water intrusion position
    xtgood=np.zeros(301, dtype=np.float64)
    if iflag==1:
        idxgood=np.zeros(301, dtype=np.float64)
        ytgood=np.zeros(301, dtype=np.float64)
        ugood=np.zeros(301, dtype=np.float64)
        for j in range(301):
            flag=0
            i=0
            while flag==0:
                idx=j*701+i
                if u_array[idx]<phi and u_array[idx+1]>phi:
                    xtgood[k1]=coor[idx+1][0]
                    ytgood[k1]=coor[idx+1][1]
                    idxgood[k1]=idx+1
                    k1=k1+1
                    flag=1
                i=i+1
        # Plots
        l1x=np.linspace(0,7000)
        l1y=0.0*l1x+1950.0
        l2x=np.linspace(0,6000)
        l2y=0.0*l2x+1200.0
        l3y=np.linspace(0,1200)
        l3x=0.0*l3y+2650.0
        l4y=np.linspace(1200,1950)
        l4x=0.0*l4y+6000.0
        if flagplot==1:
            if fplot==0:
                plt.ion()
                plt.axis('scaled')
                plt.plot(G1x,G1y,'r.',G2x,G2y,'r.',G3x,G3y,'r.',G4x,G4y,'r.',G5x,G5y,'r.',3920.0, 960.0,'bs',2640.0, 1560.0,'g^',4860.0, 1580.0,'bo',3340.0, 2200.0,'rs',4620.0, 2460.0,'r^',xtgood, ytgood, 'b.',l1x,l1y,'m-',l2x,l2y,'m-',l3x,l3y,'m-',l4x,l4y,'m-')
                plt.text(1000, 2300, "$K_1$=25", size=10)
                plt.text(1000, 1500, "$K_2$=35", size=10)
                plt.text(5000, 600, "$K_3$=50", size=10)
                plt.text(1000, 600, "$K_4$=75", size=10)
                circle1=plt.Circle((3920.0, 960.0),400.0,color='b',fill=False)
                circle2=plt.Circle((2640.0,1560.0),400.0,color='b',fill=False)
                circle3=plt.Circle((4860.0,1580.0),400.0,color='b',fill=False)
                circle4=plt.Circle((3340.0,2200.0),400.0,color='b',fill=False)
                circle5=plt.Circle((4620.0,2460.0),400.0,color='b',fill=False)
                fig = plt.gcf()
                fig.gca().add_artist(circle1)
                fig.gca().add_artist(circle2)
                fig.gca().add_artist(circle3)
                fig.gca().add_artist(circle4)
                fig.gca().add_artist(circle5)
                plt.axis([0, 7000, 0, 3000])
                plt.draw()
                plt.pause(0.000001)
                plt.clf()
            elif fplot==1:
                plt.ion()
                plt.axis('scaled')
                plt.plot(G1x,G1y,'r.',G2x,G2y,'r.',G3x,G3y,'r.',G4x,G4y,'r.',G5x,G5y,'r.',3920.0, 960.0,'bs',2640.0, 1560.0,'g^',4860.0, 1580.0,'bo',3340.0, 2200.0,'rs',4620.0, 2460.0,'r^',xtgood, ytgood, 'r.',l1x,l1y,'m-',l2x,l2y,'m-',l3x,l3y,'m-',l4x,l4y,'m-')
                plt.text(1000, 2300, "$K_1$=25", size=10)
                plt.text(1000, 1500, "$K_2$=35", size=10)
                plt.text(5000,  600, "$K_3$=50", size=10)
                plt.text(1000,  600, "$K_4$=75", size=10)
                circle1=plt.Circle((3920.0, 960.0),400.0,color='b',fill=False)
                circle2=plt.Circle((2640.0,1560.0),400.0,color='b',fill=False)
                circle3=plt.Circle((4860.0,1580.0),400.0,color='b',fill=False)
                circle4=plt.Circle((3340.0,2200.0),400.0,color='b',fill=False)
                circle5=plt.Circle((4620.0,2460.0),400.0,color='b',fill=False)
                fig = plt.gcf()
                fig.gca().add_artist(circle1)
                fig.gca().add_artist(circle2)
                fig.gca().add_artist(circle3)
                fig.gca().add_artist(circle4)
                fig.gca().add_artist(circle5)
                plt.axis([0, 7000, 0, 3000])
                plt.draw()
                plt.pause(0.000001)
                plt.clf()
            else:
                plt.ioff()
                plt.axis('scaled')
                plt.plot(G1x,G1y,'r.',G2x,G2y,'r.',G3x,G3y,'r.',G4x,G4y,'r.',G5x,G5y,'r.',3920.0, 960.0,'bs',2640.0, 1560.0,'g^',4860.0, 1580.0,'bo',3340.0, 2200.0,'rs',4620.0, 2460.0,'r^',xtgood, ytgood, 'g.',l1x,l1y,'m-',l2x,l2y,'m-',l3x,l3y,'m-',l4x,l4y,'m-')
                plt.text(1000, 2300, "$K_1$=25", size=10)
                plt.text(1000, 1500, "$K_2$=35", size=10)
                plt.text(5000, 600, "$K_3$=50", size=10)
                plt.text(1000, 600, "$K_4$=75", size=10)
                circle1=plt.Circle((3920.0, 960.0),400.0,color='b',fill=False)
                circle2=plt.Circle((2640.0,1560.0),400.0,color='b',fill=False)
                circle3=plt.Circle((4860.0,1580.0),400.0,color='b',fill=False)
                circle4=plt.Circle((3340.0,2200.0),400.0,color='b',fill=False)
                circle5=plt.Circle((4620.0,2460.0),400.0,color='b',fill=False)
                fig = plt.gcf()
                fig.gca().add_artist(circle1)
                fig.gca().add_artist(circle2)
                fig.gca().add_artist(circle3)
                fig.gca().add_artist(circle4)
                fig.gca().add_artist(circle5)
                plt.axis([0, 7000, 0, 3000])
                plt.show()
                plot(u)
                interactive()

    return xtgood,DW
